"""
This module contains the repository abstractions used by pyrrhic.
"""
import json
from dataclasses import dataclass
from pathlib import Path
from typing import Generator

from pyrrhic.crypto import keys
from pyrrhic.repo import index, pack, snapshot


@dataclass(frozen=True)
class Repository:
    """Class that is used for high level access of Repository."""

    repository: Path
    masterkey: keys.MasterKey

    def get_index(self, index_prefix: str = "", glob=True) -> index.Index:
        return index.Index(self.masterkey, self.repository, index_prefix, glob)

    def get_snapshot(self, snapshot_prefix: str = "") -> Generator[snapshot.Snapshot, None, None]:
        return snapshot.get_snapshot(self.masterkey, self.repository, snapshot_prefix)

    def get_config(self) -> dict:
        ct = Path(self.repository / "config").read_bytes()
        pt = keys.decrypt_mac(self.masterkey, ct)
        return json.loads(pt)

    def get_pack(self, pack_id: str) -> pack.Pack:
        return pack.Pack(self.repository, self.masterkey, pack_id)


def get_masterkey(repository: Path, password: str) -> keys.MasterKey:
    if not password:
        raise ValueError("Please specify password")
    if not repository:
        raise ValueError("Please specify repository location")

    keys_path = repository / "keys"
    for kf in keys_path.iterdir():
        try:
            return keys.get_masterkey(kf, password.encode("utf8"))
        except ValueError as err:
            last_err = err
    raise last_err



from base64 import b64decode
from pathlib import Path

from pyrrhic.crypto.keys import (
    Mac,
    MasterKey,
    WrappedKey,
    load_key,
)
from pyrrhic.repo.repository import Repository, get_masterkey

import pytest

REPO_BASE = "restic_test_repositories"
KEY = "98f9e68226bf15a8e9616632df7c9df543e255b388bfca1cde0218009b77cdeb"
KEYPATH = Path(f"{REPO_BASE}/restic_test_repository/keys/{KEY}")  # noqa: E501
REPO = Path(REPO_BASE) / "restic_test_repository"
BROKEN_REPO = Path(REPO_BASE) / "restic_broken_test_repository"


def test_load_key():
    key = load_key(KEYPATH)
    assert type(key) == WrappedKey


def test_get_masterkey():
    masterkey = get_masterkey(REPO, "password")
    assert masterkey == MasterKey(
        mac=Mac(
            k=b64decode("aSbwRFL8rIOOxL4W+mAW1w=="),
            r=b64decode("hQYBDSD/JwpU8XMDIJmAAg=="),
        ),
        encryption=b64decode("Te0IPiu0wvEtr2+J59McgTrjCp/ynVxC/mmM9mX/t+E="),
    )


def test_get_masterkey_with_invalid_password():
    with pytest.raises(ValueError):
        get_masterkey(REPO, "password2")


def test_config_with_invalid_mac():
    masterkey = get_masterkey(REPO, "password")
    repo = Repository(BROKEN_REPO, masterkey)
    with pytest.raises(ValueError, match="ciphertext verification failed"):
        repo.get_config()



from pathlib import Path

import pyrrhic.cli.state
from pyrrhic.repo.repository import Repository, get_masterkey

import pytest


def test_cat_masterkey(capfd):
    with pytest.raises(ValueError):
        pyrrhic.cli.state.repository = Repository(
            Path("restic_test_repositories/restic_test_repository"),
            get_masterkey(Path("restic_test_repositories/restic_test_repository"), "invalid!"),
        )
        assert capfd.readouterr().out == "Invalid Password"



import re
from datetime import datetime
from typing import Any, Type

DATETIME_TRAILING_NS_RE = re.compile(r"(\.[0-9]{6})[0-9]*\+")


def datetime_from_restic(s: str) -> datetime:
    return datetime.fromisoformat(DATETIME_TRAILING_NS_RE.sub(r"\1+", s))


class resticdatetime(datetime):
    @classmethod
    def fromresticformat(cls, date_string):
        """Construct a datetime from the restic datetime str (strip nanoseconds)."""
        date_string = DATETIME_TRAILING_NS_RE.sub(r"\1+", date_string)
        return cls.fromisoformat(date_string)


def resticdatetime_enc_hook(obj: Any) -> Any:
    if isinstance(obj, resticdatetime):
        return obj.isoformat()
    else:
        raise TypeError(f"Objects of type {type(obj)} are not supported")


def resticdatetime_dec_hook(type: Type, obj: Any) -> Any:
    if type is resticdatetime:
        return resticdatetime.fromresticformat(obj)
    else:
        raise TypeError(f"Objects of type {type} are not supported")



import hashlib
from pathlib import Path

from pyrrhic.cli import state
from pyrrhic.cli.restore import restore
from pyrrhic.repo.repository import Repository, get_masterkey

REPO_BASE = Path("restic_test_repositories")
SNAPSHOT_PREFIX = "fb56c7b6"

state.repository = Repository(
    Path("restic_test_repositories/restic_test_repository"),
    get_masterkey(Path("restic_test_repositories/restic_test_repository"), "password"),
)

RESTORE_FILES = [
    (Path("usr/share/cracklib/cracklib.magic"), "c4b2b3034acf5b35b60a8de27c7ac33f54c6b4ea"),
    (Path("usr/share/cracklib/cracklib-small"), "5f97502ab12eac2e3aa869d00a13af41a7f585e6"),
    (Path("usr/share/cracklib/pw_dict.hwm"), "5dfc5fa9b8fec7eff807ede3561f4b2cdca17277"),
    (Path("usr/share/cracklib/pw_dict.pwd"), "20973efecb1e0239800a6e2437fa3390d3fd415c"),
    (Path("usr/share/cracklib/pw_dict.pwi"), "398936961dff2e5f710723c36e67edca92943284"),
]


def test_restore(tmp_path):
    restore(SNAPSHOT_PREFIX, target=tmp_path)
    for snapshot_path, sha1sum in RESTORE_FILES:
        assert hashlib.sha1((tmp_path / snapshot_path).read_bytes()).hexdigest() == sha1sum



from pathlib import Path

from pyrrhic.crypto.keys import get_masterkey
from pyrrhic.repo.snapshot import Snapshot, get_snapshot

import pytest

REPO_BASE = Path("restic_test_repositories")
SNAPSHOT_ID = "dd62b535d10bd8f24440cc300a868d6bf2f472859f1218883b0a6faca364c10c"
KEY_ID = "98f9e68226bf15a8e9616632df7c9df543e255b388bfca1cde0218009b77cdeb"
KEY_PATH = Path(f"{REPO_BASE}/restic_test_repository/keys/{KEY_ID}")


@pytest.fixture
def masterkey():
    return get_masterkey(KEY_PATH, b"password")


def test_load_snapshot(masterkey):
    snapshot = get_snapshot(masterkey, REPO_BASE / "restic_test_repository", SNAPSHOT_ID)
    assert type(next(snapshot, None)) == Snapshot


def test_load_snapshot_prefix(masterkey):
    snapshots = get_snapshot(
        masterkey,
        REPO_BASE / "restic_test_repository",
        SNAPSHOT_ID,
    )
    snapshot = next(snapshots, None)
    assert snapshot
    assert next(snapshots, None) is None


def test_load_snapshot_invalid_prefix(masterkey):
    snapshot = next(get_snapshot(masterkey, REPO_BASE / "restic_test_repository", "invalid"), None)
    assert snapshot is None



import operator
import stat
from dataclasses import dataclass

import pyrrhic.cli.state as state
from pyrrhic.repo.tree import Node, Tree, get_tree
from pyrrhic.util import datetime_from_restic

from rich import print
from rich.table import Table


import typer

_MODE_STAT = {
    "dir": stat.S_IFDIR,
    "file": stat.S_IFREG,
    "symlink": stat.S_IFLNK,
    "dev": stat.S_IFBLK,
    "chardev": stat.S_IFCHR,
    "socket": stat.S_IFSOCK,
    "fifo": stat.S_IFIFO,
}


def _print_long(node: Node, path: str, table: Table) -> None:
    mtime_str = f"{datetime_from_restic(node.mtime):%Y-%m-%d %H:%M:%S}"  # rich compatible datetime str
    if node.type == "symlink":
        name = f"{node.name} -> {node.linktarget}"
    else:
        name = node.name
    mode = stat.filemode(stat.S_IMODE(node.mode) | _MODE_STAT.get(node.type, 0))
    table.add_row(f"{mode}", f"{node.uid}", f"{node.gid}", f"{node.size}", f"{mtime_str}", f"{path}/{name}")


@dataclass(frozen=True)
class PathNode:
    "Represent a Node located at path"
    path: str  # prefix path
    node: Node


def _ls_breadth_first(tree: Tree, table: Table | None = None) -> None:
    pathnodes = [PathNode(f"/{node.name}", node) for node in tree.nodes]
    while pathnodes:
        pleafes = [pnode for pnode in pathnodes if not pnode.node.subtree]
        pathnodes = [pnode for pnode in pathnodes if pnode.node.subtree]  # FIXME: traverses 2 times
        for pleaf in pleafes:
            if table:
                _print_long(pleaf.node, pleaf.path, table)
            else:
                print(f"{pleaf.path}/{pleaf.node.name}")
        if pathnodes:
            pnode = pathnodes.pop()
            tree = get_tree(state.repository, pnode.node.subtree or "0123")  # just make mypy happy?
            # FIXME: Order by packref to improve performance
            pathnodes = [*[PathNode(f"{pnode.path}/{node.name}", node) for node in tree.nodes], *pathnodes]


def ls(snapshot_prefix: str, long: bool = typer.Option(False, "--long", "-l", help="Use a long listing (Unix ls -l)")):
    "List files in a snapshot"
    state.repository.get_snapshot(snapshot_prefix)
    if snapshot_prefix == "latest":
        snapshots = iter(sorted(state.repository.get_snapshot(), key=operator.attrgetter("time"), reverse=True)[:1])
    else:
        snapshots = state.repository.get_snapshot(snapshot_prefix)
    snapshot = next(snapshots, None)
    if not snapshot:
        raise ValueError(f"Index: {snapshot_prefix} not found")
    if next(snapshots, None):
        raise ValueError(f"Prefix {snapshot_prefix} matches multiple snapshots")
    tree = get_tree(state.repository, snapshot.tree)
    if long:
        table = Table("mode", "user", "group", "size", "date", "filename", highlight=True)
        _ls_breadth_first(tree, table)
        print(table)
    else:
        _ls_breadth_first(tree)



import logging
import sys
from enum import Enum
from pathlib import Path

import pyrrhic
import pyrrhic.cli.cat as cat
import pyrrhic.cli.snapshots as snapshots
import pyrrhic.cli.state
from pyrrhic.cli.ls import ls
from pyrrhic.cli.restore import restore
from pyrrhic.cli.util import catch_exception
from pyrrhic.repo.repository import Repository, get_masterkey

from rich.logging import RichHandler

import typer

app: typer.Typer = typer.Typer(add_completion=False)
app.add_typer(cat.app, name="cat", help="🐈 Print internal objects to stdout")
app.command()(snapshots.snapshots)
app.command()(ls)
app.command()(restore)


class LogLevel(str, Enum):
    debug = "DEBUG"
    info = "INFO"
    warn = "WARN"
    error = "ERROR"

    def __str__(self):
        return self.value


@app.command()
def version():
    """Return version of pyrric application"""
    print(pyrrhic.__version__)


@app.callback()
@catch_exception(OSError, exit_code=2)
@catch_exception(ValueError, exit_code=1)
def global_options(
    loglevel: LogLevel = typer.Option(LogLevel.error, case_sensitive=False),
    repo: Path = typer.Option(None, "--repo", "-r", help="repository for subcommands ", envvar="RESTIC_REPOSITORY"),
    password: str = typer.Option(
        None,
        help="repository password",
        envvar="RESTIC_PASSWORD",
    ),
    password_file: Path = typer.Option(None, "--password-file", "-p", help="file to read the repository password from", envvar="RESTIC_PASSWORD_FILE"),
):
    logging.basicConfig(level=str(loglevel), format="%(name)s: %(message)s", datefmt="[%X]", handlers=[RichHandler()])
    if password_file:
        if password:
            print("password and password-file are mutually exclusive", file=sys.stderr)
            raise typer.Exit(code=1)
        password = Path(password_file).read_text().strip()
    if not password:
        password = typer.prompt("repository password", hide_input=True)

    masterkey = get_masterkey(repo, password)
    pyrrhic.cli.state.repository = Repository(repo, masterkey)


if __name__ == "__main__":
    app()



import hashlib
import json
from ast import literal_eval
from pathlib import Path

import pyrrhic.cli.state
from pyrrhic.cli.cat import config, index, masterkey, pack, snapshot
from pyrrhic.cli.ls import ls
from pyrrhic.cli.snapshots import snapshots
from pyrrhic.repo.repository import Repository, get_masterkey

pyrrhic.cli.state.repository = Repository(
    Path("restic_test_repositories/restic_test_repository"),
    get_masterkey(Path("restic_test_repositories/restic_test_repository"), "password"),
)


def test_cat_masterkey(capfd):
    masterkey()
    assert literal_eval(capfd.readouterr().out) == {
        "encrypt": "Te0IPiu0wvEtr2+J59McgTrjCp/ynVxC/mmM9mX/t+E=",
        "mac": {"k": "aSbwRFL8rIOOxL4W+mAW1w==", "r": "hQYBDSD/JwpU8XMDIJmAAg=="},
    }


def test_cat_config(capfd):
    config()
    assert literal_eval(capfd.readouterr().out) == {
        "chunker_polynomial": "3833148ec41f8d",
        "id": "2ec6792a9c75a017ec4665a5e7733f45f8bffb67c7d0f3c9ec6cc96e58c6386b",
        "version": 1,
    }


def test_cat_snapshot(capfd):
    snapshot("dd62b535d10bd8f24440cc300a868d6bf2f472859f1218883b0a6faca364c10c")
    out = capfd.readouterr().out
    assert json.loads(out) == {
        "time": "2022-07-19T21:52:28.692251+02:00",
        "tree": "a5dbcc77f63f5dd4f4c67c988aba4a19817aaa9d6c34a6021236a5d40ce653e1",
        "paths": ["/home/juergen/shared/python/pyrrhic/test"],
        "hostname": "shaun",
        "username": "juergen",
        "id": "dd62b535d10bd8f24440cc300a868d6bf2f472859f1218883b0a6faca364c10c",
        "uid": 1000,
        "gid": 1000,
        "excludes": None,
        "tags": None,
    }


def test_cat_index(capfd):
    index("0de57faa699ec0450ddbafb789e165b4e1a3dbe3a09b071075f09ebbfbd6f4b2")
    out = capfd.readouterr().out
    assert "PackRef" in out
    assert (
        """'511d1c632ccad135d5407157154eccc17fcfaf501ad252b231c7ce41175473b9': PackRef(
        id='46771395523ccd6dda16694f0ce775f9508a4c3e4527c385f55d8efafa36807f'"""
        in out
    )


def test_cat_pack(capfdbinary):
    sha = "4b24375f07a164e995d06303bfc26f79f94127e4e5c6e1c476495bbee0af7ccc"
    pack(sha, False)
    assert hashlib.sha256(capfdbinary.readouterr().out).hexdigest() == sha


def test_cat_pack_header(capfd):
    sha = "4b24375f07a164e995d06303bfc26f79f94127e4e5c6e1c476495bbee0af7ccc"
    pack(sha, True)
    assert literal_eval(capfd.readouterr().out) == [
        {"id": "82deec86c5611cb5ae02b967e49d7aeaca50a732432bd1a5923787bd5d0fbf80", "length": 1643, "offset": 0},
        {"id": "e20d6400fbd4e602c79c8bab98a88726865b168838cc8107d560da10f19b2ff8", "length": 1467, "offset": 1643},
        {"id": "a5dbcc77f63f5dd4f4c67c988aba4a19817aaa9d6c34a6021236a5d40ce653e1", "length": 413, "offset": 3110},
    ]


def test_snapshots(capfd):
    snapshots()
    output = capfd.readouterr().out
    assert "ID" in output


def test_ls(capfd):
    "Returns a list of paths"
    ls("latest", False)
    lines = capfd.readouterr().out.splitlines()
    for s in lines:
        path = Path(s)
        assert path.is_absolute()



import operator
import os
import shutil
import sys

import msgspec

import pyrrhic.cli.state
from pyrrhic.cli.util import catch_exception
from pyrrhic.util import resticdatetime_enc_hook

from rich import print, print_json

import typer


app: typer.Typer = typer.Typer(add_completion=False)


@app.command()
@catch_exception(ValueError, exit_code=1)
@catch_exception(FileNotFoundError, exit_code=2)
def masterkey():
    """Return masterkey JSON to stdout"""
    state = pyrrhic.cli.state
    print(state.repository.masterkey.restic_json())


@app.command()
@catch_exception(ValueError, exit_code=1)
@catch_exception(FileNotFoundError, exit_code=2)
def config():
    """Return config JSON to stdout"""
    state = pyrrhic.cli.state
    config = state.repository.get_config()
    print(config)


@app.command()
@catch_exception(ValueError, exit_code=1)
@catch_exception(FileNotFoundError, exit_code=2)
def index(index_id: str):
    """Return index JSON to stdout"""
    state = pyrrhic.cli.state
    index = state.repository.get_index(index_id)
    print(index.index)


@app.command()
@catch_exception(ValueError, exit_code=1)
@catch_exception(FileNotFoundError, exit_code=2)
def snapshot(snapshot_id: str):
    """Return snapshot JSON to stdout"""
    state = pyrrhic.cli.state
    enc = msgspec.json.Encoder(enc_hook=resticdatetime_enc_hook)
    if snapshot_id == "latest":
        snapshots = iter(sorted(pyrrhic.cli.state.repository.get_snapshot(), key=operator.attrgetter("time"), reverse=True)[:1])
    else:
        snapshots = state.repository.get_snapshot(snapshot_id)
    if (snapshot := next(snapshots, None)) and next(snapshots, None) is None:
        print_json(enc.encode(snapshot).decode("utf-8"))
    else:
        raise ValueError(f"Invalid Index: {snapshot_id}")


@app.command()
@catch_exception(ValueError, exit_code=1)
@catch_exception(FileNotFoundError, exit_code=2)
def pack(pack_id: str, header: bool = typer.Option(False, "--header", help="Output parsed pack header")):
    """Return pack to stdout"""
    state = pyrrhic.cli.state
    pack = state.repository.get_pack(pack_id)
    if header:
        print(list(pack.get_blob_index()))
    else:
        with os.fdopen(sys.stdout.fileno(), "wb", closefd=False) as stdout, open(pack.path, "rb") as pack_fd:
            shutil.copyfileobj(pack_fd, stdout)



import hashlib
from typing import Optional

import msgspec

from pyrrhic.crypto.keys import decrypt_mac
from pyrrhic.repo.repository import Repository


class Node(msgspec.Struct):
    name: str
    type: str
    mode: int
    mtime: str
    atime: str
    ctime: str
    uid: int
    gid: int
    size: int = 0
    user: str
    inode: int
    content: Optional[list[str]] = None
    linktarget: Optional[str] = None
    subtree: Optional[str] = None


class Tree(msgspec.Struct):
    nodes: list[Node]


def get_node_blob(repo: Repository, blob_id: str) -> bytes:
    index = repo.get_index()
    packref = index.get_packref(blob_id)
    blob = packref.blob
    with open(repo.repository / "data" / packref.id[:2] / packref.id, "rb") as f:
        f.seek(blob.offset)
        buffer = f.read(blob.length)
        plaintext = decrypt_mac(repo.masterkey, buffer)
        if hashlib.sha256(plaintext).hexdigest() != blob.id:
            raise ValueError(f"Invalid hash for blob {blob.id}")
        return plaintext
    raise ValueError(f"Can't find blob id {blob_id}")


def get_tree(repo: Repository, tree_id: str) -> Tree:
    plaintext_blob = get_node_blob(repo, tree_id)
    return msgspec.json.decode(plaintext_blob, type=Tree)



import operator

import pyrrhic.cli.state

from rich import print
from rich.table import Table


def snapshots():
    "List all snapshots"
    table = Table("ID", "datetime", "hostname", "tags", "paths", highlight=True)
    for s in sorted(pyrrhic.cli.state.repository.get_snapshot(), key=operator.attrgetter("time")):
        table.add_row(s.id[:6], f"{s.time:%c}", s.hostname, ", ".join(s.tags or []), ", ".join(s.paths or []))
    print(table)



from dataclasses import dataclass
from functools import cache
from pathlib import Path

import msgspec

from pyrrhic.crypto.keys import MasterKey, decrypt_mac

from rich.console import Console
from rich.progress import track


class Blob(msgspec.Struct):
    id: str
    type: str
    offset: int
    length: int


class BlobList(msgspec.Struct):
    id: str
    blobs: list[Blob]


class RecPackList(msgspec.Struct):
    packs: list[BlobList]


@dataclass(frozen=True)
class PackRef:
    id: str
    blob: Blob


@cache
def _get_index(key: MasterKey, repo_path: Path, index_prefix: str, glob: bool) -> dict[str, PackRef]:
    if glob:
        paths = (repo_path / "index").glob(f"{index_prefix}*")
    else:
        paths = (repo_path / "index" / name for name in [index_prefix])
    if Console().is_terminal:  # FIXME: Should be configurable
        paths = track(list(paths), "Loading index")

    dec = msgspec.json.Decoder(type=RecPackList)
    d = {
        blob.id: PackRef(packs.id, blob)
        for index_path in paths
        for packs in dec.decode(decrypt_mac(key, index_path.read_bytes())).packs
        for blob in packs.blobs
    }
    return d


class Index:
    "Internal Index representation"
    index: dict[str, PackRef]

    def __init__(self, key: MasterKey, repo_path: Path, index_prefix: str, glob=True):
        self.index = _get_index(key, repo_path, index_prefix, glob)

    def get_packref(self, blob_id) -> PackRef:
        if packref := self.index.get(blob_id):
            return packref
        raise ValueError(f"Invalid blob id {blob_id}")



import os
from functools import wraps

import typer


def get_file_by_prefix(dir: str, prefix: str) -> list[str]:
    return [file for file in os.listdir(dir) if file == str or file.startswith(prefix)]


def catch_exception(which_exception, exit_code=1):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except which_exception as e:
                typer.secho(e, err=True, fg=typer.colors.RED)
                raise typer.Exit(code=exit_code)

        return wrapper

    return decorator



"""
This module contains the crypto primitives used by restic.
"""

import json
from base64 import b64decode, b64encode
from dataclasses import dataclass
from pathlib import Path

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import poly1305
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.kdf.scrypt import Scrypt

import msgspec

from pyrrhic.util import resticdatetime, resticdatetime_dec_hook

# mask for key, (cf. http://cr.yp.to/mac/poly1305-20050329.pdf)
_POLY1305KEYMASK = b"\xff\xff\xff\x0f\xfc\xff\xff\x0f\xfc\xff\xff\x0f\xfc\xff\xff\x0f"


class WrappedKey(msgspec.Struct):
    """Class that contain all data that is needed to derive the
    repository's master encryption and message authentication keys
    from a user's password."""

    hostname: str
    username: str
    kdf: str
    N: int
    r: int
    p: int
    created: resticdatetime
    data: bytes
    salt: bytes


@dataclass(frozen=True)
class Mac:
    """Class that holdes the Poly1305-AES parameters"""

    k: bytes
    r: bytes


@dataclass(frozen=True)
class MasterKey:
    """Class that holds encryption and message authentication keys for a
    repository."""

    mac: Mac
    encryption: bytes

    def restic_json(self):
        """Return restic representation of Masterkey"""
        return {
            "mac": {
                "r": b64encode(self.mac.r).decode(),
                "k": b64encode(self.mac.k).decode(),
            },
            "encrypt": b64encode(self.encryption).decode(),
        }

    def __hash__(self):
        return hash(self.encryption + self.mac.k + self.mac.r)


def load_key(key_path: Path) -> WrappedKey:
    return msgspec.json.decode(key_path.read_bytes(), type=WrappedKey, dec_hook=resticdatetime_dec_hook)


def _poly1305_validate(nonce: bytes, k: bytes, r: bytes, message: bytes, mac: bytes) -> bool:
    r = bytes([(r & m) for r, m in zip(r, _POLY1305KEYMASK)])
    cipher = Cipher(algorithms.AES(k), modes.ECB())
    encryptor = cipher.encryptor()
    aes_ciphertext = encryptor.update(nonce) + encryptor.finalize()
    poly1305_key = r + aes_ciphertext
    p = poly1305.Poly1305(poly1305_key)
    p.update(message)
    return p.finalize() == mac


def _decrypt(aes_key: bytes, nonce: bytes, ciphertext: bytes) -> bytes:
    cipher = Cipher(algorithms.AES(aes_key), modes.CTR(nonce))
    decryptor = cipher.decryptor()
    return decryptor.update(ciphertext) + decryptor.finalize()


def decrypt_mac(key: MasterKey, restic_blob: bytes) -> bytes:
    """Decrypt 'IV || CIPHERTEXT || MAC' bytes, validate mac and
    return plaintext bytes"""
    nonce = restic_blob[:16]
    mac = restic_blob[-16:]
    ciphertext = restic_blob[16:-16]
    if not _poly1305_validate(nonce, key.mac.k, key.mac.r, ciphertext, mac):
        raise ValueError("ciphertext verification failed")
    return _decrypt(key.encryption, nonce, ciphertext)


def get_masterkey(path: Path, password: bytes) -> MasterKey:
    wrapped_key = load_key(path)
    kdf = Scrypt(
        wrapped_key.salt,
        64,
        wrapped_key.N,
        wrapped_key.r,
        wrapped_key.p,
        backend=default_backend,
    )
    derived_key = kdf.derive(password)
    poly_k = derived_key[32:48]
    poly_r = derived_key[48:]
    nonce = wrapped_key.data[0:16]
    message = wrapped_key.data[16:-16]
    mac = wrapped_key.data[-16:]
    if not _poly1305_validate(nonce, poly_k, poly_r, message, mac):
        raise ValueError("Invalid Password")
    j = json.loads(_decrypt(derived_key[:32], nonce, message))
    encryption = b64decode(j["encrypt"])
    r = b64decode(j["mac"]["r"])
    k = b64decode(j["mac"]["k"])
    return MasterKey(Mac(k=k, r=r), encryption)



from pathlib import Path

from pyrrhic.repo.index import Index
from pyrrhic.repo.repository import Repository, get_masterkey

REPO_BASE = "restic_test_repositories"
INDEX_ID = "0de57faa699ec0450ddbafb789e165b4e1a3dbe3a09b071075f09ebbfbd6f4b2"


def test_load_index():
    repo = Repository(
        Path(REPO_BASE) / "restic_test_repository",
        get_masterkey(Path(REPO_BASE) / "restic_test_repository", "password"),
    )
    indexes = repo.get_index(INDEX_ID)
    assert type(indexes) == Index


def test_load_indexes():
    repo = Repository(
        Path(REPO_BASE) / "restic_test_repository",
        get_masterkey(Path(REPO_BASE) / "restic_test_repository", "password"),
    )
    indexes = repo.get_index()
    assert type(indexes) == Index



import os
from dataclasses import dataclass
from pathlib import Path
from struct import unpack

from pyrrhic.crypto.keys import MasterKey, decrypt_mac


@dataclass(frozen=True)
class Blob:
    pack_id: str
    offset: int
    length: int
    hash: bytes


@dataclass(frozen=True)
class DataBlob(Blob):
    pass


@dataclass(frozen=True)
class TreeBlob(Blob):
    pass


@dataclass(frozen=True)
class CompressedDataBlob(Blob):
    length_uncompressed: int


@dataclass(frozen=True)
class CompressedTreeBlob:
    length_uncompressed: int


class Pack:
    def __init__(self, repo_path: Path, key: MasterKey, pack_id: str):
        "Load Pack at path."
        self.pack_id = pack_id
        self.path = repo_path / "data" / pack_id[:2] / pack_id
        with open(self.path, "rb") as f:
            f.seek(-4, os.SEEK_END)
            buffer = f.read(4)
            header_length = unpack("<I", buffer)[0]
            f.seek(-4 - header_length, os.SEEK_END)
            buffer = f.read(header_length)
            self.header = decrypt_mac(key, buffer)

    def get_blobs(self):
        header = self.header
        offset = 0
        while len(header):
            match header[0]:
                case 0:
                    length_encrypted = unpack("<I", header[1:5])[0]
                    yield DataBlob(self.pack_id, offset, length_encrypted, header[5 : 5 + 32])
                    header = header[(5 + 32) :]
                case 1:
                    length_encrypted = unpack("<I", header[1:5])[0]
                    yield TreeBlob(self.pack_id, offset, length_encrypted, header[5 : 5 + 32])
                    header = header[(5 + 32) :]
                case 2:
                    length_encrypted, length_plaintext = unpack("<I<I", header[1:9])  # FIXME: Wrong in Restic-specification, shoud be length_uncompressed?
                    yield CompressedDataBlob(self.pack_id, offset, length_encrypted, header[9 : 9 + 32], length_plaintext)
                    header = header[9 + 32 :]
                case 3:
                    length_encrypted, length_plaintext = unpack("<I<I", header[1:9])
                    yield CompressedTreeBlob(self.pack_id, offset, length_encrypted, header[9 : 9 + 32], length_plaintext)
                    header = header[9 + 32 :]
                case _:
                    raise ValueError(f"Invalid Tag: {header[0]}")
            offset += length_encrypted

    def get_blob_index(self):
        "Return blobs in index format"
        return ({"id": blob.hash.hex(), "offset": blob.offset, "length": blob.length} for blob in self.get_blobs())



import operator
import stat
from logging import info, warn
from pathlib import Path

import pyrrhic.cli.state as state
from pyrrhic.cli.util import catch_exception
from pyrrhic.repo.tree import Tree, get_node_blob, get_tree


def _restore_recursive(tree: Tree, target: Path):
    for node in tree.nodes:
        abs_path = target / node.name
        mode = stat.S_IMODE(node.mode)
        match node.type:
            case "file":
                # FIXME: Create temporary unique file
                abs_path.touch(mode, exist_ok=False)  # FIXME: UID/GID
                if node.content:  # possible empty file
                    info(f"Restoring {abs_path}: {len(node.content)} blobs")
                    with open(abs_path, "wb") as f:
                        for content_id in node.content:
                            f.write(get_node_blob(state.repository, content_id))

            case "dir":
                info(f"Restoring dir {abs_path}")
                abs_path.mkdir(mode)
                if node.subtree:
                    _restore_recursive(get_tree(state.repository, node.subtree), abs_path)
            case _:
                warn(f"{node.name}: {node.type} not implemented")


@catch_exception(OSError, exit_code=2)
def restore(snapshot_prefix: str, target: Path, help="Restore data from a snapshot"):
    state.repository.get_snapshot(snapshot_prefix)
    if snapshot_prefix == "latest":  # FIXME: Duplicated code (ls command)
        snapshots = iter(sorted(state.repository.get_snapshot(), key=operator.attrgetter("time"), reverse=True)[:1])
    else:
        snapshots = state.repository.get_snapshot(snapshot_prefix)
    snapshot = next(snapshots, None)
    if not snapshot:
        raise ValueError(f"Index: {snapshot_prefix} not found")
    if next(snapshots, None):
        raise ValueError(f"Prefix {snapshot_prefix} matches multiple snapshots")
    tree = get_tree(state.repository, snapshot.tree)
    _restore_recursive(tree, target)



from pathlib import Path
from typing import Generator, List, Optional

import msgspec

from pyrrhic.crypto.keys import MasterKey, decrypt_mac
from pyrrhic.util import resticdatetime, resticdatetime_dec_hook


class Snapshot(msgspec.Struct):
    id: Optional[str] = None  # filled by pyrrhic
    time: resticdatetime
    tree: str
    paths: List[str]
    hostname: str
    username: str
    uid: Optional[int] = None  # Undocumented https://restic.readthedocs.io/en/stable/100_references.html#repository-format
    gid: Optional[int] = None
    excludes: Optional[List[str]] = None
    tags: Optional[List[str]] = None


def get_snapshot(key: MasterKey, repo_path: Path, snapshot_prefix: str) -> Generator[Snapshot, None, None]:
    dec = msgspec.json.Decoder(Snapshot, dec_hook=resticdatetime_dec_hook)
    for snapshot_path in (repo_path / "snapshots").glob(f"{snapshot_prefix}*"):
        snapshot = dec.decode(decrypt_mac(key, snapshot_path.read_bytes()))
        snapshot.id = snapshot_path.name
        yield snapshot



"""
Nox automation tasks for pyrrhic
"""

import os
from pathlib import Path

import nox

# GitHub Actions
ON_CI = bool(os.getenv("CI"))

# Git info
DEFAULT_BRANCH = "master"

# Python to use for non-test sessions
DEFAULT_PYTHON: str = "3.10"

# Global project stuff
PROJECT_ROOT = Path(__file__).parent.resolve()

SOURCE_FILES = (
    "noxfile.py",
    "pyrrhic",
    "test",
)


@nox.session(python=DEFAULT_PYTHON)
def dev(session: nox.Session) -> None:
    """
    Sets up a python dev environment for the project if one doesn't already exist.
    """
    session.run("python", "-m", "venv", os.path.join(PROJECT_ROOT, ".venv"))
    session.run(
        "flit",
        "install",
        "--symlink",
        "--python",
        os.path.join(PROJECT_ROOT, ".venv", "bin", "python"),
        external=True,
        silent=True,
    )


@nox.session(python=["3.10"])
def test(session: nox.Session) -> None:
    """
    Runs the test suite.
    """
    session.install(
        ".",
        "pytest",
        "pytest-cov",
        "flake8",
    )
    session.run("flake8", *SOURCE_FILES)
    session.run("pytest", "-v", "-v", "--cov=pyrrhic", "--cov-branch")


@nox.session(python=DEFAULT_PYTHON)
def coverage(session: nox.Session):
    """Upload coverage data."""
    session.install("coverage[toml]", "codecov")
    session.run("coverage", "xml", "--fail-under=0")
    session.run("codecov", *session.posargs)


@nox.session(python=DEFAULT_PYTHON)
def mypy(session: nox.Session) -> None:
    """Type-check using mypy."""
    args = session.posargs or ["pyrrhic", "test"]
    session.install("mypy")
    session.run("mypy", "--ignore-missing-imports", "--show-error-codes", *args)


@nox.session(python=DEFAULT_PYTHON)
def lint(session: nox.Session) -> None:
    """Lint using flake8."""
    args = session.posargs or ["pyrrhic", "test"]
    session.install(
        "flake8",
        "flake8-black",
        "flake8-import-order",
    )
    session.run("flake8", *args)


@nox.session(python=DEFAULT_PYTHON)
def sphinx(session: nox.Session) -> None:
    """Generate Sphinx documentation from source files."""
    session.install(".", "sphinx")
    session.run(
        "sphinx-apidoc",
        "--force",
        "--implicit-namespaces",
        "--module-first",
        "--separate",
        "-o",
        "docs/reference/",
        "pyrrhic",
    )
    session.run("sphinx-build", "-W", "--keep-going", "-b", "html", "docs/", "docs/_build/")



from pathlib import Path

from pyrrhic.repo.index import Index
from pyrrhic.repo.pack import Pack
from pyrrhic.repo.repository import Repository, get_masterkey

import pytest

REPO_BASE = "restic_test_repositories"
INDEX_ID = "0de57faa699ec0450ddbafb789e165b4e1a3dbe3a09b071075f09ebbfbd6f4b2"
TEST_REPO = Path("./restic_test_repositories/restic_test_repository")


@pytest.fixture
def masterkey():
    return get_masterkey(TEST_REPO, "password")


def test_load_pack(masterkey):
    p = Pack(TEST_REPO, masterkey, "46771395523ccd6dda16694f0ce775f9508a4c3e4527c385f55d8efafa36807f")
    assert p


def test_index_matches_packs(masterkey):
    repo = Repository(Path(REPO_BASE) / "restic_test_repository", masterkey)
    index = repo.get_index(INDEX_ID)
    assert type(index) == Index
    for blob_id, packref in index.index.items():
        p = Pack(repo.repository, masterkey, packref.id)
        pack_blobs = p.get_blob_index()
        assert (matching_blob := next((blob for blob in pack_blobs if blob["id"] == blob_id)))
        assert matching_blob["offset"] == packref.blob.offset
        assert matching_blob["length"] == packref.blob.length
